# -*- coding: utf8 -*-
from ..dulwich.object_store import BaseObjectStore


class NullObjectStore(BaseObjectStore):
    def add_objects(self, objects):
        pass

    def __iter__(self):
        pass

    def contains_packed(self, sha):
        pass

    @property
    def packs(self):
        pass

    def contains_loose(self, sha):
        pass

    def add_object(self, obj):
        pass

    def get_raw(self, name):
        raise KeyError()



# -*- coding: utf8 -*-
import copy
import logging
from abc import ABCMeta, abstractmethod

import datetime
import six
import sys

from ..connection_mixin import ConnectionMixin


class MetadataOperationError(Exception):
    pass


class MetadataTypeNotSupported(MetadataOperationError):
    pass


class MetadataFieldNotFound(MetadataOperationError):
    def __init__(self, field):
        self._field = field

    @property
    def field(self):
        return self._field

    def __str__(self):
        return 'Metadata field %s not found' % self.field


def unicode_dict_to_str(d):
    if sys.version_info >= (3, 0):
        return d

    for key, val in d.items():
        if not isinstance(val, unicode):
            continue

        d[key] = val.encode('utf8')

    return d


class BaseMetadataDB(ConnectionMixin):
    __metaclass__ = ABCMeta

    STAGING_COMMIT = 'staging'
    DEFAULT_SEED = 1337
    DEFAULT_SPLIT = {'train': 0.6, 'test': 0.2, 'validation': 0.2}

    def __init__(self, connection, version_ts_lookup=None, **kwargs):
        super(BaseMetadataDB, self).__init__(connection)
        self.__rnd = None
        self.__prev_table_info = None
        self.__create_table_if_needed()
        self.__version_ts_lookup = version_ts_lookup

    @abstractmethod
    def _create_table(self):
        """

        :return:
        """

    @abstractmethod
    def _add_missing_columns(self, data_object):
        """

        :param data_object:
        :return:
        """

    @abstractmethod
    def _add_data(self, flatten_data_list):
        """

        :param flatten_data_list:
        :return:
        """

    @abstractmethod
    def get_data_for_commit(self, sha, commit_sha):
        """

        :param sha:
        :param commit_sha:
        :return:
        """

    @abstractmethod
    def get_all_data(self, sha):
        """

        :param sha:
        :return:
        """

    @abstractmethod
    def begin_commit(self, commit_sha, tree_id, ts):
        """

        :param commit_sha:
        :param tree_id:
        :return:
        """

    @abstractmethod
    def end_commit(self):
        """

        :param commit_sha:
        :param tree_id:
        :return:
        """

    @abstractmethod
    def _query(self, sql_vars, select_fields, where, max_results, start_index):
        """

        :param sql_vars:
        :param select_fields:
        :param where:
        :return:
        """

    @abstractmethod
    def _query_head_data(self, sha_list):
        """

        :param sha_list:
        :return: list
        """

        return []

    def __create_table_if_needed(self):
        if self._connection.read_only:
            return

        self._create_table()

    def add_data(self, data):
        if not data:
            logging.debug('no data provided')
            return

        if not isinstance(data, list):
            data = [data]

        logging.debug('add data (total: %s)', len(data))

        data_list = []
        for sha, data_object in data:
            data_object_clone = copy.deepcopy(data_object)

            self._add_missing_columns(data_object)

            data_object_clone['@sha'] = sha
            data_object_clone['@commit_sha'] = self.STAGING_COMMIT
            data_object_clone['@version'] = 2 ** 32 - 1
            data_object_clone['@hash'] = None

            data_list.append(data_object_clone)

        self._add_data(data_list)

    def explicit_query(self, query_text):
        from ..scam import QueryParser, tree_to_sql_parts

        tree = QueryParser().parse_query(query_text)
        sql_vars, _ = tree_to_sql_parts(tree, self._connection.create_sql_helper())

        sql_vars = sql_vars or {}

        if 'seed' not in sql_vars:
            query_text += ' @seed:%s' % self.DEFAULT_SEED

        return query_text

    def query(self, query_text, max_results, start_index):
        from ..scam import QueryParser, tree_to_sql_parts, get_split_vars, ParseError

        from pyparsing import ParseException

        if query_text is not None and not isinstance(query_text, str):
            query_text = query_text.encode('utf8')

        try:
            tree = QueryParser().parse_query(query_text)
            sql_vars, where = tree_to_sql_parts(tree, self._connection.create_sql_helper())
        except ParseException:
            raise ParseError()

        select_fields = ['*']

        sql_vars = sql_vars or {}

        if 'seed' not in sql_vars:
            sql_vars['seed'] = self.DEFAULT_SEED

        if 'phase_train_start' not in sql_vars:
            sql_vars.update(get_split_vars(self.DEFAULT_SPLIT))

        sql_vars['version_ts'] = datetime.datetime.utcnow()
        if 'version' in sql_vars:
            version = sql_vars['version']
            if self.__version_ts_lookup is not None:
                sql_vars['version_ts'] = self.__version_ts_lookup(version)

        if 'sample_percentile' not in sql_vars:
            sql_vars['sample_percentile'] = 0.0
            sql_vars['sample'] = 1.0

        return self._query(sql_vars, select_fields, where, max_results, start_index)

    @classmethod
    def fill_in_vars(cls, query_sql, sql_vars):
        var_keys = sorted(sql_vars.keys(), reverse=True)
        for var_name in var_keys:
            var_value = sql_vars[var_name]
            query_sql = query_sql.replace('$' + var_name, str(var_value))

        return query_sql

    def get_head_data(self, sha_list):
        if isinstance(sha_list, six.string_types):
            sha_list = [sha_list]

        return self._query_head_data(sha_list)

    def _return_all_result_from_query(self, query_sql, batch_size=1000):
        from flatten_json import unflatten

        with self._connection.get_cursor() as c:
            c.execute(query_sql)

            while True:
                results = c.fetchmany(batch_size)
                if not results:
                    break

                for result in results:
                    yield unflatten(unicode_dict_to_str(result), separator='.')



# -*- coding: utf8 -*-
import contextlib
import os
import re
import shutil
import textwrap

import click

# The trick where we squash multiple spaces into one optimizes for readability.
# It shows the control flow, while keeping the actual shell code to one line
# (important for ease-of-RC-file-management).
_SOURCE_LINE_SH = ' '.join(filter(None, """\
if [ -f '{rc_path}' ]; then \
    source '{rc_path}'; \
fi
""".split(' ')))
_SOURCE_LINE_FISH = ' '.join(filter(None, """\
if [ -f '{rc_path}' ]; \
    if type source > /dev/null; \
       source '{rc_path}'; \
    else; \
       . '{rc_path}'; \
    end; \
end
""".split(' ')))


def _get_rc_contents(comment, rc_path, rc_contents, pattern=None, source_line=_SOURCE_LINE_SH):
    """Generates the RC file contents with new comment and `source rc_path` lines.

    Args:
      comment: The shell comment string that precedes the source line.
      rc_path: The path of the rc file to source.
      rc_contents: The current contents.
      pattern: A regex pattern that matches comment, None for exact match on
        comment.
      source_line: str, the template for sourcing a file in the shell being
        updated ('{rc_path}' will be substituted with the file to source)

    Returns:
      The comment and `source rc_path` lines to be inserted into a shell rc file.
    """
    if not pattern:
        pattern = re.escape(comment)
    # This pattern handles all three variants that we have injected in user RC
    # files. All have the same sentinel comment line followed by:
    #   1. a single 'source ...' line
    #   2. a 3 line if-fi (a bug because this pattern was previously incorrect)
    #   3. finally a single if-fi line.
    # If you touch this code ONLY INJECT ONE LINE AFTER THE SENTINEL COMMENT LINE.
    #
    # At some point we can drop the alternate patterns and only search for the
    # sentinel comment line and assume the next line is ours too (that was the
    # original intent before the 3-line form was added).
    subre = re.compile(
        '\n' + pattern + '\n('
                         "source '.*'"
                         '|'
                         'if .*; then\n  source .*\nfi'
                         '|'
                         'if .*; then source .*; fi'
                         '|'
                         'if .*; if type source .*; end'
                         ')\n', re.MULTILINE)

    # script checks that the rc_path currently exists before sourcing the file
    line = ('\n{comment}\n' + source_line).format(comment=comment, rc_path=rc_path)
    filtered_contents = subre.sub('', rc_contents)
    rc_contents = '{filtered_contents}{line}'.format(filtered_contents=filtered_contents, line=line)

    return rc_contents


TEXTWRAP = textwrap.TextWrapper(replace_whitespace=False,
                                drop_whitespace=False,
                                break_on_hyphens=False)


@contextlib.contextmanager
def _narrow_wrap(narrow_by):
    """Temporarily narrows the global wrapper."""
    TEXTWRAP.width -= narrow_by
    yield TEXTWRAP
    TEXTWRAP.width += narrow_by


def format_required_user_action(s):
    """Formats an action a user must initiate to complete a command.

    Some actions can't be prompted or initiated by gcloud itself, but they must
    be completed to accomplish the task requested of gcloud; the canonical example
    is that after installation or update, the user must restart their shell for
    all aspects of the update to take effect. Unlike most console output, such
    instructions need to be highlighted in some way. Using this function ensures
    that all such instances are highlighted the *same* way.

    Args:
      s: str, The message to format. It shouldn't begin or end with newlines.

    Returns:
      str, The formatted message. This should be printed starting on its own
        line, and followed by a newline.
    """
    with _narrow_wrap(4) as wrapper:
        return '\n==> ' + '\n==> '.join(wrapper.wrap(s)) + '\n'


class RcUpdater(object):
    """Updates the RC file completion and PATH code injection."""

    def __init__(self, shell, completion_path, rc_path):
        self.rc_path = rc_path
        self.completion = completion_path
        self.shell = shell

    def _CompletionExists(self):
        return os.path.exists(self.completion)

    def _GetSourceLine(self):
        if self.shell == 'fish':
            return _SOURCE_LINE_FISH
        else:
            return _SOURCE_LINE_SH

    def update(self):
        #  Check whether RC file is a file and store its contents.
        if os.path.isfile(self.rc_path):
            with open(self.rc_path) as rc_file:
                rc_contents = rc_file.read()
                original_rc_contents = rc_contents
        elif os.path.exists(self.rc_path):
            click.echo(
                '[{rc_path}] exists and is not a file, so it cannot be updated.'.format(rc_path=self.rc_path),
                err=True)
            return
        else:
            rc_contents = ''
            original_rc_contents = ''

        if self._CompletionExists():
            rc_contents = _get_rc_contents(
                '# The next line enables shell command completion for mali (MissingLink.AI command line).',
                self.completion, rc_contents, source_line=self._GetSourceLine(),
                pattern=('# The next line enables [a-z][a-z]*'
                         ' command completion for mali \(MissingLink.AI command line\).'))

        if rc_contents == original_rc_contents:
            click.echo('No changes necessary for [{rc}].'.format(rc=self.rc_path))
            return

        if os.path.exists(self.rc_path):
            rc_backup = self.rc_path + '.backup'
            click.echo('Backing up [{rc}] to [{backup}].'.format(rc=self.rc_path, backup=rc_backup))
            shutil.copyfile(self.rc_path, rc_backup)

        with open(self.rc_path, 'w') as rc_file:
            rc_file.write(rc_contents)

        click.echo('[{rc_path}] has been updated.'.format(rc_path=self.rc_path))
        click.echo(format_required_user_action('Start a new shell for the changes to take effect.'))


def __get_shell_rc_file_name_from_shell(shell):
    if shell == 'ksh':
        return os.environ.get('ENV', None) or '.kshrc'

    if shell == 'fish':
        return os.path.join('.config', 'fish', 'config.fish')

    if shell != 'bash':
        return '.{shell}rc'.format(shell=shell)

    return None


def __get_shell_rc_file_name_from_platform():
    from sys import platform

    if platform == "linux" or platform == "linux2":
        return '.bashrc'

    if platform == "darwin":
        return '.bash_profile'

    if platform == "win32":
        return '.profile'

    return '.bashrc'


def _get_shell_rc_file_name(shell):
    rc_name = __get_shell_rc_file_name_from_shell(shell)

    if rc_name is None:
        rc_name = __get_shell_rc_file_name_from_platform()

    return rc_name


def rc_updater(shell, completion_path):
    rc_path = os.path.join(os.path.expanduser('~'), _get_shell_rc_file_name(shell))

    # Check the rc_path for a better hint at the user preferred shell.
    RcUpdater(shell, completion_path, rc_path).update()



# -*- coding: utf8 -*-
import json
import logging

import requests
import sys

import time
from requests import HTTPError
from six.moves.urllib import parse

from .eprint import eprint
from .config import get_prefix_section, Config

BASE_URL_PATH = '_ah/api/missinglink/v1/'


def __urljoin(*args):
    base = args[0]
    for u in args[1:]:
        base = parse.urljoin(base, u)

    return base


def __api_call(config, url, http_method, data):
    id_token = config.id_token

    for _ in range(3):
        headers = {'Authorization': 'Bearer {}'.format(id_token)}
        r = http_method(url, headers=headers, json=data)

        if r.status_code == 401:
            try:
                id_token = update_token(config)
                continue
            except HTTPError:
                eprint('Authorization failed, try running "mali auth init" again')
                sys.exit(1)

        r.raise_for_status()

        return r.json()

    eprint('failed to refresh the token, rerun auth init again')
    sys.exit(1)


def _handle_async_api(config, http_method, method_url, data, retry):
    if data is not None:
        data['async'] = True

    result = _handle_sync_api(config, http_method, method_url, data, retry)
    token = result['token']

    while True:
        result = _handle_sync_api(config, requests.get, 'data_volumes/tasks/' + token)
        if result.get('failed'):
            raise Exception('Internal Server Error ')

        if result.get('finished'):
            return json.loads(result['results']) if 'results' in result else None

        time.sleep(1.0)


def _handle_sync_api(config, http_method, method_url, data=None, retry=None):
    if config.id_token is None:
        eprint('Please run: "mali auth init" to setup authorization')
        sys.exit(1)

    url = __urljoin(config.api_host, BASE_URL_PATH, method_url)

    def api_call_with_retry(current_config, current_url, current_http_method, current_data):
        return retry.call(__api_call, current_config, current_url, current_http_method, current_data)

    f = __api_call if retry is None else api_call_with_retry

    try:
        return f(config, url, http_method, data)
    except requests.exceptions.HTTPError as ex:
        try:
            error_message = ex.response.json().get('error', {}).get('message')
        except ValueError:
            error_message = None

        if error_message is None:
            error_message = str(ex)

        eprint('\n' + error_message)
        sys.exit(1)


def handle_api(ctx_or_config, http_method, method_url, data=None, retry=None, async=False):
    config = ctx_or_config if isinstance(ctx_or_config, Config) else ctx_or_config.config

    if async:
        return _handle_async_api(config, http_method, method_url, data, retry)

    return _handle_sync_api(config, http_method, method_url, data, retry)


def build_auth0_url(auth0):
    return '{}.auth0.com'.format(auth0)


def _should_retry_auth0(exception):
    logging.debug('got retry exception (auth0) %s', exception)

    error_codes_to_retries = [
        429,  # Too many requests
    ]

    return isinstance(exception, requests.exceptions.HTTPError) and exception.response.status_code in error_codes_to_retries


def update_token(config):
    from retrying import retry

    @retry(retry_on_exception=_should_retry_auth0)
    def with_retry():
        r = requests.post('https://{}/delegation'.format(build_auth0_url(config.auth0)), json={
            'client_id': config.client_id,
            'grant_type': "urn:ietf:params:oauth:grant-type:jwt-bearer",
            'scope': 'openid offline_access user_external_id org orgs email picture name given_name user_metadata',
            'refresh_token': config.refresh_token,
        })

        r.raise_for_status()

        data = r.json()

        config.set(get_prefix_section(config.config_prefix, 'token'), 'id_token', data['id_token'])
        config.save()

        return data['id_token']

    return with_retry()


def default_api_retry(stop_max_attempt_number=None):
    from retrying import Retrying

    def retry_if_retry_possible_error(exception):
        logging.debug('got retry exception (api) %s', exception)

        return True

    return Retrying(
        retry_on_exception=retry_if_retry_possible_error,
        wait_exponential_multiplier=50,
        wait_exponential_max=5000,
        stop_max_attempt_number=stop_max_attempt_number)



# -*- coding: utf8 -*-
import csv
import logging

import requests
import six

from ..api import default_api_retry
from .base_db_index import BaseMLIndex
from ..backend_mixin import BackendMixin


class BackendMLIndex(BackendMixin, BaseMLIndex):
    def __init__(self, connection, config, handle_api):
        super(BackendMLIndex, self).__init__(connection, config, handle_api)

    def _create_table_if_needed(self):
        pass

    def set_entries(self, entries):
        if not entries:
            return

        if isinstance(entries, six.string_types):
            self.set_entries_from_url(entries)
        else:
            self.set_entries_from_list(entries)

    def set_entries_from_url(self, file_name):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage' % self._volume_id

            msg = {
                'index_url': file_name,
            }

            self._handle_api(self._config, session.post, url, msg, retry=default_api_retry(), async=True)

    def set_entries_from_list(self, entries):
        rows = []
        for name, sha, ctime, mtime, mode, uid, gid, size, url in self._decode_entries(entries):
            row = {
                'name': name,
                'sha': sha,
                'ctime': ctime,
                'mtime': mtime,
                'mode': mode,
                'uid': uid,
                'gid': gid,
                'size': size,
                'url': url,
            }

            rows.append(row)

        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage' % self._volume_id
            msg = {
                'entries': rows,
            }

            self._handle_api(self._config, session.post, url, msg, retry=default_api_retry())

    class ChangeSetIter(object):
        def __init__(self, data, data_type=None):
            self.__data = data
            data_type = data_type or 'csv'

            if data_type.lower() == 'csv':
                self.__reader = csv.DictReader(data)

        def __iter__(self):
            return self

        def next(self):
            data = next(self.__reader)
            return data['name'], data['op']

        def __next__(self):
            return self.next()

    def get_changeset(self, data):
        if not data:
            logging.debug('no data provided')
            return

        logging.debug('add data %s', data)

        def decode_iter(gen):
            for line in gen:
                if not isinstance(line, six.string_types):
                    line = line.decode()

                yield line

        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage' % self._volume_id

            msg = {
                'index_url': data,
                'dry_mode': True,
            }

            result = self._handle_api(self._config, session.post, url, msg, retry=default_api_retry())

            r = requests.get(result['change_set_url'], stream=True)
            r.raise_for_status()

            return self.ChangeSetIter(decode_iter(r.iter_lines()))

    def begin_commit(self, commit_sha, tree_id, ts):
        raise NotImplementedError(self.begin_commit)

    def end_commit(self):
        raise NotImplementedError(self.end_commit)



# -*- coding: utf8 -*-
import collections
import importlib
import json
import logging
import threading
from .metadata_files import MetadataFiles

try:
    import cPickle as pickle
except ImportError:
    import pickle

import random
import string
import tempfile
from contextlib import closing
import sys
import os
from uuid import uuid4
import datetime
import requests
from six import wraps

from .api import default_api_retry
from .config import Config
from .data_volume import with_repo
from .json_utils import multi_line_json_from_data, normalize_item, dict_normalize
from . import MetadataOperationError
from .multi_process_control import get_multi_process_control
from .path_utils import get_batch_of_files_from_paths, safe_make_dirs, DestPathEnum, enumerate_paths_with_info
from .print_status import PrintStatus
from .eprint import eprint


def wrap_exceptions(fn):
    @wraps(fn)
    def decorated(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except AssertionError:
            raise
        except Exception as ex:
            eprint('\n' + str(ex))
            sys.exit(1)

    return decorated


epoch = datetime.datetime.utcfromtimestamp(0)


class InvalidJsonFile(Exception):
    def __init__(self, filename):
        self.filename = filename


def _download_entity(config_init_dict, volume_id, metadata):
    return DataSync.download_entity(config_init_dict, volume_id, metadata)


def chop_microseconds(delta):
    return delta - datetime.timedelta(microseconds=delta.microseconds)


def status_with_timing(message, callback):
    start_time = datetime.datetime.utcnow()
    status_line = PrintStatus()
    with closing(status_line):
        status_line.print_status(message)
        result = callback()
        total_time = datetime.datetime.utcnow() - start_time
        total_time = chop_microseconds(total_time)
        status_line.print_status('%s Done (%s)' % (message, total_time))

    return result


def disk_result(resume_token, method, result_callback=None):
    def wrap_f(f):
        def wrap(*args, **kwargs):
            import tempfile

            def loader():
                cache_result = pickle.load(cache_file)
                if result_callback is not None:
                    result_callback(cache_result)

                return cache_result

            name = '{func_name}_{python_version}'.format(func_name=method, python_version=sys.hexversion)
            full_path = os.path.join(tempfile.gettempdir(), 'ml_cache', resume_token, name)

            try:
                with open(full_path, 'rb') as cache_file:
                    return status_with_timing('Loading cache file', loader)
            except IOError:
                pass
            except Exception as ex:
                logging.debug('failed to load %s %s', full_path, ex)

            result = f(*args, **kwargs)

            full_path_tmp = full_path + '.tmp'
            try:
                safe_make_dirs(os.path.dirname(full_path_tmp))
                with open(full_path_tmp, 'wb') as cache_file:
                    pickle.dump(result, cache_file, pickle.HIGHEST_PROTOCOL)
                    os.rename(full_path_tmp, full_path)
            except Exception as ex:
                logging.info('failed to save %s %s', full_path_tmp, ex)

            return result

        return wrap

    return wrap_f


class DiskStorage(object):
    def __init__(self, dest_pattern, save_meta):
        self.__dest_pattern = dest_pattern
        self.__save_meta = save_meta

    def close(self):
        pass

    def add_item(self, metadata, data):
        full_filename = DestPathEnum.get_full_path(self.__dest_pattern, metadata)
        safe_make_dirs(os.path.dirname(full_filename))

        if self.__save_meta:
            item_meta = {key: val for key, val in metadata.items() if not key.startswith('@')}
            with open(full_filename + '.metadata.json', 'w') as f:
                json.dump(item_meta, f)

        with open(full_filename, 'wb') as f:
            f.write(data)

    def has_item(self, metadata):
        full_filename = DestPathEnum.get_full_path(self.__dest_pattern, metadata)
        return os.path.isfile(full_filename)

    @property
    def storage_params(self):
        return {
            'dest_pattern': self.__dest_pattern,
            'save_meta': self.__save_meta,
        }

    @classmethod
    def init_from_config(cls, dest_pattern, save_meta, **kwargs):
        return cls(dest_pattern, save_meta)


class DataSync(object):
    local_data = threading.local()

    resume_prefix = 'resume-'

    def __init__(self, ctx, repo, no_progressbar, resume_token=None):
        self.__ctx = ctx
        self.__repo = repo
        self.__no_progressbar = no_progressbar

        self.__resume_token = resume_token or self.generate_resume_token()

        if not self.__resume_token.startswith(self.resume_prefix):
            self.__resume_token = self.resume_prefix + self.__resume_token

    @property
    def repo(self):
        return self.__repo

    @property
    def resume_token(self):
        return self.__resume_token

    def __upload_file_for_processing(self, file_obj, file_description):
        from .gcs_utils import do_upload
        from tqdm import tqdm

        volume_id = self.__repo.data_volume_config.volume_id

        data_object_name = '%s/temp/%s_%s.json' % (volume_id, file_description, uuid4().hex)

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=volume_id)

        headers = {'Content-Type': 'application/json'}

        msg = {
            'methods': 'PUT',
            'paths': [data_object_name],
            'content_type': 'application/json',
            'temp': True,
        }

        result = self.__ctx.obj.handle_api(self.__ctx.obj, requests.post, url, msg, retry=default_api_retry())

        put_url = result['put'][0]

        gcs_auth = self.__repo.data_volume_config.object_store_config.get('auth')

        def update_bar(c):
            bar.update(c)

        logging.debug('uploading %s', data_object_name)
        with tqdm(total=file_obj.tell(), desc="Uploading {}".format(file_description), unit=' KB', ncols=80,
                  disable=self.__no_progressbar, unit_scale=True) as bar:
            do_upload(gcs_auth, None, data_object_name, file_obj, headers, head_url=None, put_url=put_url,
                      callback=update_bar)

        return data_object_name

    def __process_index(self, object_name):
        index = self.__repo.open_index()

        change_set = status_with_timing('Server process index', lambda: index.get_changeset(data='gs://' + object_name))

        total_modify_files = 0
        total_new_files = 0

        files_to_upload = tempfile.TemporaryFile('w+')

        for name, op in change_set:
            if op == 'm':
                total_modify_files += 1
            elif op == 'i':
                total_new_files += 1
            else:
                continue

            data = name + '\n'

            files_to_upload.write(data)

        total_files_to_upload = total_modify_files + total_new_files
        files_to_upload.seek(0)

        return files_to_upload, total_files_to_upload

    def upload_and_update_index(self, index_entities_file):
        object_name = self.__upload_file_for_processing(index_entities_file, 'Index')

        files_to_upload, total_files_to_upload = self.__process_index(object_name)

        return files_to_upload, total_files_to_upload

    def upload_and_update_metadata(self, file_obj):
        object_name = self.__upload_file_for_processing(file_obj, 'Metadata')

        status_with_timing('Server process metadata', lambda: self.__repo.metadata.add_data(data='gs://' + object_name))

    def __append_index_file(self, file_info, combined_index_files):
        file_name = file_info['path']

        mtime = file_info['mtime']
        ctime = file_info.get('ctime', mtime)

        mtime = (mtime - epoch).total_seconds()
        ctime = (ctime - epoch).total_seconds()

        if file_info['sys'] == 'local':
            rel_path = os.path.relpath(file_name, self.__repo.data_path)

        else:  # file_info['sys'] == 's3':
            rel_path = file_name

        params = {
            'ctime': ctime,
            'mtime': mtime,
            'size': file_info['size'],
            'sha': file_info['sha'],
            'mode': 0
        }

        if 'url' in file_info:
            params['url'] = file_info['url']

        multi_line_json_from_data({rel_path: params}, combined_index_files, key_name='name')

    def __create_combined_index_and_metadata(self, data_path):
        import humanize
        status_line = PrintStatus()

        def on_result(result):
            data_files_info, metadata_files_list = result

            total_files_size = sum(x['size'] for x in data_files_info.values())

            status_line.print_status(
                'Total files {:,} ({}) (metadata found: {:,})', len(data_files_info),
                humanize.naturalsize(total_files_size), len(metadata_files_list))

        @disk_result(self.resume_token, '__create_combined_index_and_metadata', on_result)
        def with_disk_result():
            total_files = 0
            total_files_size = 0
            total_metadata = 0

            data_files_list = set()
            metadata_files_list = set()
            data_files_info = {}

            with closing(status_line):
                for file_info in enumerate_paths_with_info(data_path):
                    rel_file_name = self.repo.rel_path(file_info['path'])
                    total_files_size += file_info['size']

                    if rel_file_name.endswith(MetadataFiles.metadata_ext):
                        total_metadata += 1
                        metadata_files_list.add(rel_file_name)
                    else:
                        total_files += 1
                        data_files_list.add(rel_file_name)
                        data_files_info[rel_file_name] = file_info

                    status_line.print_status(
                        'Total files {:,} ({}) (metadata found: {:,})', total_files,
                        humanize.naturalsize(total_files_size), total_metadata)

            return data_files_info, metadata_files_list

        return with_disk_result()

    def create_combined_index_and_metadata(self, data_path):
        @disk_result(self.resume_token, 'create_combined_index_and_metadata')
        def with_disk_result():
            data_files_info, metadata_files_list = self.__create_combined_index_and_metadata(data_path)

            files_metadata = MetadataFiles.load_all_metadata(
                self.__repo, data_files_info, metadata_files_list, self.__no_progressbar)

            return files_metadata, data_files_info

        return with_disk_result()

    @wrap_exceptions
    def upload_in_batches(self, files_info, total_files=None, callback=None):
        total_files = total_files or len(files_info)
        batch_size = max(min(total_files // 100, 250), 250)  # FIXME: hardcoded

        for files_info_batch in get_batch_of_files_from_paths(files_info, batch_size):
            self.__repo.stage(files_info_batch, callback=callback)

    @classmethod
    def __object_from_data(cls, data, creator):
        data_key = json.dumps(data, sort_keys=True)

        try:
            return cls.local_data.__data_sync_objects[data_key]
        except KeyError:
            cls.local_data.__data_sync_objects[data_key] = creator(data)
        except AttributeError:
            cls.local_data.__data_sync_objects = {data_key: creator(data)}

        return cls.local_data.__data_sync_objects[data_key]

    @classmethod
    def download_entity(cls, config_init_dict, volume_id, metadata):
        config = cls.__object_from_data(config_init_dict, lambda current_data: Config(**current_data))

        def get_item_data():
            try:
                _, current_data = repo.object_store.get_raw(metadata)
                return current_data
            except requests.exceptions.HTTPError as ex:
                if ex.response.status_code == 404:
                    return

                raise

        def get_storage(storage_class):
            module_name, class_name = storage_class.rsplit('.', 1)
            m = importlib.import_module(module_name)
            return getattr(m, class_name)

        storage_config = config.items('storage')

        def add_new_storage(current_data):
            return get_storage(current_data['class']).init_from_config(**current_data)

        storage = cls.__object_from_data(storage_config, add_new_storage)

        with with_repo(config, volume_id, read_only=True) as repo:
            with closing(storage):
                if storage.has_item(metadata):
                    return

                data = get_item_data()
                storage.add_item(metadata, data)

    def download_all(self, query, root_folder, dest_pattern, batch_size, processes, save_meta=True):
        multi_process_control = get_multi_process_control(processes)
        return self.__download_all(multi_process_control, query, root_folder, dest_pattern, batch_size, save_meta=save_meta)

    def __download_results(self, process_control, storage, results, callback=None):
        volume_id = self.__repo.data_volume_config.volume_id

        def wrap_callback_func(current_item):
            def wrapper(_):
                if callback is not None:
                    callback(current_item)

            return wrapper

        def fullname(o):
            return o.__module__ + "." + o.__class__.__name__

        config_init_dict = self.__ctx.obj.config.init_dict
        storage_params = storage.storage_params
        storage_params['class'] = fullname(storage)
        config_init_dict['storage'] = storage_params

        for normalized_item in results:
            task_async = process_control.execute(
                _download_entity,
                args=(config_init_dict, volume_id, normalized_item),
                callback=wrap_callback_func(normalized_item))

            yield task_async, normalized_item

    class DownloadIterResults(collections.Iterator):
        def __init__(self, repo, query, batch_size):
            self.__query = query
            self.__batch_size = batch_size
            self.__repo = repo
            self.__query_index = 0
            self.__total_data_points = None

        def __get_next_results(self):
            current_results, current_total_data_points = self.__repo.metadata.query(
                self.__query, max_results=self.__batch_size, start_index=self.__query_index)

            return [normalize_item(metadata) for metadata in current_results], current_total_data_points

        def __len__(self):
            return self.__total_data_points

        def __next__(self):
            results, self.__total_data_points = self.__get_next_results()
            self.__query_index += len(results)

            if len(results) == 0:
                raise StopIteration()

            return results

        next = __next__

    def create_download_iter(self, query, batch_size):
        return self.DownloadIterResults(self.__repo, query, batch_size)

    def download_items(self, metadata_items, storage, multi_process_control):
        if len(metadata_items) == 0:
            return

        futures = []
        for async_result, normalized_item in self.__download_results(multi_process_control, storage, metadata_items):
            futures.append((async_result, normalized_item))

        for async_result, normalized_item in futures:
            async_result.wait()

    def __download_all(self, multi_process_control, query, root_folder, dest_pattern, batch_size, save_meta=True):
        from tqdm import tqdm

        def handle_item(item):
            file_name = DestPathEnum.get_full_path(dest_pattern, item)
            rel_path = os.path.relpath(file_name, root_folder)
            phase = item.get('phase', 'all')
            phase_meta.setdefault(phase, {})[rel_path] = item

            bar.update()

        download_iter = self.create_download_iter(query, batch_size)

        results = next(download_iter)
        phase_meta = {}

        storage = DiskStorage(dest_pattern, save_meta)

        with tqdm(total=len(download_iter), desc="Downloading", unit=' data point', ncols=80, disable=self.__no_progressbar) as bar:
            with closing(multi_process_control):
                try:
                    while True:
                        for _ in self.__download_results(multi_process_control, storage, results, handle_item):
                            pass  # no need to wait for the feature becouse the closing context above will wait for the entire pool

                        if len(results) != batch_size:
                            break

                        results = next(download_iter)
                except MetadataOperationError as ex:
                    eprint(str(ex))
                except KeyboardInterrupt:
                    multi_process_control.terminate()
                    multi_process_control.close()

        return phase_meta

    def save_metadata(self, root_dest, metadata):
        from tqdm import tqdm

        with tqdm(total=len(metadata), desc="saving metadata", unit=' files', ncols=80, disable=self.__no_progressbar) as bar:
            for key, val in metadata.items():
                if key is None:
                    key = 'unknown'

                json_metadata_file = os.path.join(root_dest, key + '.metadata.json')

                with open(json_metadata_file, 'w') as f:
                    json.dump(val, f)

                bar.update()

    def create_index_file_from_data(self, f, files_info):
        for val in files_info.values():
            self.__append_index_file(val, f)

    @classmethod
    def create_metadata_file_from_data(cls, f, metadata_files):
        dict_normalize(metadata_files)

        for rel_file_name, file_metadata in metadata_files.items():
            multi_line_json_from_data({rel_file_name: file_metadata}, f)

    def __get_files_to_upload(self, files_info, combined_index_files):
        files_to_upload, total_files_to_upload = self.upload_and_update_index(combined_index_files)

        def switch_to_full_path(params):
            rel_path = params['path']

            params['path'] = self.repo.full_path(rel_path)

            return params

        def enum_file_names():
            for file_name in files_to_upload:
                file_name = file_name.strip()

                yield file_name

        files_to_upload = [switch_to_full_path(files_info[rel_file_name]) for rel_file_name in enum_file_names()]

        return files_to_upload

    def __upload_full_index(self, index_entities_file):
        object_name = self.__upload_file_for_processing(index_entities_file, 'Index')
        index = self.__repo.open_index()

        status_with_timing('Server process index', lambda: index.set_entries('gs://' + object_name))

    def upload_index_and_metadata(self, data_path):
        def with_disk_result():
            metadata_files, files_info = self.create_combined_index_and_metadata(data_path)

            if len(files_info) == 0:
                return []

            if len(metadata_files) > 0:
                combined_meta_files = tempfile.TemporaryFile('wb+')
                self.create_metadata_file_from_data(combined_meta_files, metadata_files)
                self.upload_and_update_metadata(combined_meta_files)

            combined_index_files = tempfile.TemporaryFile('wb+')
            self.create_index_file_from_data(combined_index_files, files_info)

            if self.repo.data_volume_config.embedded:
                return self.__get_files_to_upload(files_info, combined_index_files)

            return self.__upload_full_index(combined_index_files)

        return with_disk_result()

    @classmethod
    def generate_resume_token(cls):
        size = 6
        chars = string.digits
        result = ''.join(random.choice(chars) for _ in range(size))

        return result



# -*- coding: utf8 -*-
import logging

import os
import requests

from ..path_utils import has_moniker
from ..dulwich import objects
from ..dulwich.repo import Repo
from ..api import handle_api
from ..data_volume_config import DataVolumeConfig
from .. import BigQueryConnection, DatastoreConnection, BackendConnection
from .. import BigQueryMLIndex, BackendMLIndex
from .. import BackendMetadataDB, BigQueryMetadataDB
from .. import DatastoreRefContainer, BackendRefContainer


class MLIgnoreFilterManager(object):
    def is_ignored(self, relpath):
        return False


DEFAULT_ENCODING = 'utf-8'


def make_bytes(c):
    if not isinstance(c, bytes):
        return c.encode(DEFAULT_ENCODING)

    return c


class MlRepo(Repo):
    def __init__(self, config, repo_root, read_only=False, require_path=True, data_volume_config=None):
        if data_volume_config is None:
            general_config_path = config.config_file_abs_path if config is not None else None
            self.__data_volume_config = DataVolumeConfig(repo_root, general_config_path=general_config_path)
        else:
            self.__data_volume_config = data_volume_config

        self.__in_transactions = False
        self.__config = config
        self.__connections = {}
        self.__read_only = read_only
        self.__metadata = None

        super(MlRepo, self).__init__(repo_root, self.data_volume_config.data_path, require_path=require_path)

    @property
    def _config(self):
        return self.__config

    def close(self):
        for connection in self.__connections.values():
            connection.close()

        super(MlRepo, self).close()

    __connection_class_mapping = {
        'bq': BigQueryConnection,
        'datastore': DatastoreConnection,
    }

    def rel_path(self, full_path):
        local_path = not has_moniker(self.data_path)

        if local_path:
            return os.path.relpath(full_path, self.data_path)

        return full_path

    def full_path(self, rel_path):
        local_path = not has_moniker(self.data_path)

        if local_path:
            return os.path.join(self.data_path, rel_path)

        return self.data_path + '/' + rel_path

    def __create_connection(self, name, **kwargs):
        kwargs.update(self.data_volume_config.db_config)
        kwargs['read_only'] = self.__read_only
        kwargs['data_volume_config'] = self.data_volume_config
        kwargs['user_id'] = self._config.user_id if self._config is not None else None

        connection_class = self.__connection_class_mapping.get(self.data_volume_config.db_type, BackendConnection)

        return connection_class(**kwargs)

    def start_transactions(self):
        for connection in self.__connections.values():
            connection.start_transactions()

        self.__in_transactions = True

    def end_transactions(self):
        for connection in self.__connections.values():
            connection.end_transactions()

        self.__in_transactions = False

    def rollback_transactions(self):
        for connection in self.__connections.values():
            connection.rollback_transactions()

        self.__in_transactions = False

    def _connection_by_name(self, name, **kwargs):
        if name not in self.__connections:
            connection = self.__create_connection(name, **kwargs)

            if self.__in_transactions:
                connection.start_transactions()

            self.__connections[name] = connection

        return self.__connections[name]

    def __create_metadata(self):
        if self.data_volume_config.db_type == 'bq':
            return BigQueryMetadataDB(self._connection_by_name('metadata'))

        return BackendMetadataDB(self._connection_by_name('metadata'), self._config, handle_api)

    @property
    def metadata(self):
        if self.__metadata is None:
            self.__metadata = self.__create_metadata()

        return self.__metadata

    @property
    def data_volume_config(self):
        return self.__data_volume_config

    def open_index(self):
        if self.data_volume_config.db_type == 'bq':
            return BigQueryMLIndex(self._connection_by_name('main'))

        return BackendMLIndex(self._connection_by_name('main'), self._config, handle_api)

    def create_ref_container(self):
        if self.data_volume_config.object_store_type == 'disk':
            return super(MlRepo, self).create_ref_container()

        if self.data_volume_config.db_type == 'bq':
            return DatastoreRefContainer(self._connection_by_name('datastore'))

        return BackendRefContainer(self._connection_by_name('main'), self._config, handle_api)

    def create_object_store(self):
        from .. import GCSObjectStore, NullObjectStore, BackendGCSObjectStore

        if self.data_volume_config.object_store_type == 'disk':
            return super(MlRepo, self).create_object_store()

        if self.data_volume_config.object_store_type == 'null':
            return NullObjectStore()

        bucket_name = self.data_volume_config.object_store_config.get('bucket_name')

        processes = self.data_volume_config.object_store_config.get('processes', -1)
        use_multiprocess = self.data_volume_config.get_boolean('object_store', 'use_multiprocess', True)

        if bucket_name is None:
            return BackendGCSObjectStore(
                self._connection_by_name('backend'),
                self._config, handle_api, use_multiprocess=use_multiprocess, processes=processes)

        return GCSObjectStore(self._connection_by_name('gcs'), use_multiprocess=use_multiprocess, processes=processes)

    def get_config_stack(self):
        return DataVolumeConfig(self.repo_root)

    def get_ignore_filter_manager(self):
        return MLIgnoreFilterManager()

    def _get_user_identity(self):
        import jwt

        data = jwt.decode(self._config.id_token, verify=False) if self.__config.id_token else {}

        return '{name} <{email}>'.format(**data).encode('utf8')

    def has_change_set(self, ref='HEAD'):
        ref = ref.encode('ascii')

        try:
            ref_sha = self.refs[ref]

            head_tree_sha = self[ref_sha].tree
        except KeyError:  # in case of empty tree
            head_tree_sha = objects.Tree().id

        index = self.open_index()

        commit_id = index.get_commit_id()

        return None if commit_id == head_tree_sha else commit_id

    def remote_commit(self, message):
        msg = {
            'message': message,
        }

        url = "data_volumes/{volume_id}/commit".format(volume_id=self.data_volume_config.volume_id)

        return handle_api(self._config, requests.post, url, msg, async=True)

    def local_commit(self, message):
        tree_id = self.has_change_set()

        if not tree_id:
            return

        self.start_transactions()
        try:
            commit_hash = self.do_commit(message=make_bytes(message), tree=tree_id).decode('ascii')

            self.metadata.commit(commit_hash, tree_id)

            self.end_transactions()
        except Exception:
            logging.exception('failure doing commit, preforming rollback')
            self.rollback_transactions()
            raise

    def commit(self, message):
        if self.data_volume_config.is_local:
            return self.local_commit(message)

        return self.remote_commit(message)



# reflog.py -- Parsing and writing reflog files
# Copyright (C) 2015 Jelmer Vernooij and others.
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Utilities for reading and generating reflogs.
"""

import collections

from dulwich.objects import (
    format_timezone,
    parse_timezone,
    ZERO_SHA,
    )

Entry = collections.namedtuple(
    'Entry', ['old_sha', 'new_sha', 'committer', 'timestamp', 'timezone',
              'message'])


def format_reflog_line(old_sha, new_sha, committer, timestamp, timezone,
                       message):
    """Generate a single reflog line.

    :param old_sha: Old Commit SHA
    :param new_sha: New Commit SHA
    :param committer: Committer name and e-mail
    :param timestamp: Timestamp
    :param timezone: Timezone
    :param message: Message
    """
    if old_sha is None:
        old_sha = ZERO_SHA
    return (old_sha + b' ' + new_sha + b' ' + committer + b' ' +
            str(timestamp).encode('ascii') + b' ' +
            format_timezone(timezone) + b'\t' + message)


def parse_reflog_line(line):
    """Parse a reflog line.

    :param line: Line to parse
    :return: Tuple of (old_sha, new_sha, committer, timestamp, timezone,
        message)
    """
    (begin, message) = line.split(b'\t', 1)
    (old_sha, new_sha, rest) = begin.split(b' ', 2)
    (committer, timestamp_str, timezone_str) = rest.rsplit(b' ', 2)
    return Entry(old_sha, new_sha, committer, int(timestamp_str),
                 parse_timezone(timezone_str)[0], message)


def read_reflog(f):
    """Read reflog.

    :param f: File-like object
    :returns: Iterator over Entry objects
    """
    for l in f:
        yield parse_reflog_line(l)



# -*- coding: utf8 -*-
import json
import logging
import time

from .base_connection import BaseConnection


class BigQueryOperationError(Exception):
    pass


class BigQuerySqlHelper(object):
    @classmethod
    def escape(cls, name):
        return '`%s`' % name

    @classmethod
    def random_function_name(cls):
        return '((FARM_FINGERPRINT($field_random_generator) + POW(2, 63)) / POW(2, 64))'


class BigQueryConnection(BaseConnection):
    @classmethod
    def owner_id_as_org_id(cls, user_id):
        return user_id.replace('-', '_')

    @classmethod
    def _handle_bq_errors(cls, errors):
        if not errors:
            return

        error_msg = json.dumps(errors)

        raise BigQueryOperationError(error_msg)

    def __init__(self, data_volume_config, project, user_id, **kwargs):
        org_name = data_volume_config.org or self.owner_id_as_org_id(user_id)
        self.__table_prefix = data_volume_config.volume_id

        self.__dataset_name = self.__get_dataset_fullname(org_name)
        self.__project = project
        self.__dataset = None
        super(BigQueryConnection, self).__init__(data_volume_config, **kwargs)

    @classmethod
    def __get_dataset_fullname(cls, org_name):
        return 'data_volumes_{org}'.format(org=org_name)

    @property
    def table_prefix(self):
        return self.__table_prefix

    @property
    def project(self):
        return self.__project

    @property
    def dataset(self):
        return self.__dataset_name

    def _create_connection(self, **kwargs):
        from google.cloud import bigquery

        bq_client = bigquery.Client(project=self.__project)
        return bq_client

    def load_table_from_uri(
            self, source_uris, destination,
            job_id=None, job_id_prefix=None, job_config=None, retry=None):
        return self._native_conn.load_table_from_uri(source_uris, destination, job_id, job_id_prefix, job_config, retry)

    def extract_table(self, source, destination_uris, job_config=None, job_id=None, job_id_prefix=None, retry=None):
        from google.cloud.bigquery import DEFAULT_RETRY

        retry = retry or DEFAULT_RETRY

        return self._native_conn.extract_table(
            source, destination_uris, job_config=job_config, job_id=job_id,
            job_id_prefix=job_id_prefix, retry=retry)

    def get_table(self, table_ref):
        return self._native_conn.get_table(table_ref)

    def copy_table(self, job_name, destination, *sources):
        return self._native_conn.copy_table(
            job_name, destination, *sources)

    def query(self, query, job_config=None, job_id=None, job_id_prefix=None):
        return self._native_conn.query(query, job_config=job_config, job_id=job_id, job_id_prefix=job_id_prefix)

    def update_table(self, table, properties, retry=None):
        from google.cloud.bigquery import DEFAULT_RETRY

        retry = retry or DEFAULT_RETRY

        return self._native_conn.update_table(table, properties, retry)

    def create_table(self, table):
        logging.info('create_table %s', table.table_id)

        return _bq_retry(self._native_conn.create_table, table)

    def delete_table(self, table):
        import google.cloud.exceptions
        try:
            logging.info('delete_table: %s', table.table_id)

            self._native_conn.delete_table(table)
        except google.cloud.exceptions.NotFound:
            logging.info('Table "%s" is not found', table.table_id)

    def create_rows(self, table, rows, selected_fields=None, **kwargs):
        errors = self._native_conn.create_rows(table, rows, selected_fields=selected_fields, **kwargs)
        self._handle_bq_errors(errors)

    def list_rows(
            self, table, selected_fields=None, max_results=None,
            page_token=None, start_index=None, retry=None):

        from google.cloud.bigquery import DEFAULT_RETRY

        retry = retry or DEFAULT_RETRY
        return self._native_conn.list_rows(table, selected_fields, max_results, page_token, start_index, retry)

    def _create_cursor(self):
        import google.cloud.exceptions
        from google.cloud.bigquery import Dataset

        if self.__dataset is None:
            bq_client = self._native_conn

            dataset_ref = bq_client.dataset(self.__dataset_name)
            dataset = Dataset(dataset_ref)

            if not self.read_only:
                try:
                    _bq_retry(bq_client.create_dataset, dataset)
                except google.cloud.exceptions.Conflict:
                    pass

            self.__dataset = dataset

        return self.__dataset

    def _commit(self):
        pass

    def _rollback(self):
        pass

    def create_sql_helper(self):
        return BigQuerySqlHelper()

    def delete_tables(self, prefixes=None):
        import google.cloud.exceptions

        if not prefixes:
            prefixes = ['']
        elif not isinstance(prefixes, (list, tuple)):
            prefixes = [prefixes]

        logging.info("delete table with prefixes = '%s' in dataset '%s'" % (prefixes, self.__dataset_name))

        bq_client = self._native_conn
        dataset_ref = bq_client.dataset(self.__dataset_name)

        try:
            for table_ref in bq_client.list_dataset_tables(dataset_ref):
                for prefix in prefixes:
                    if not table_ref.table_id.startswith(prefix):
                        continue

                    self.delete_table(table_ref)
        except google.cloud.exceptions.NotFound:
            # This will catch list_dataset_tables errors when the dataset is not found
            pass


INITIAL_DELAY = 1.0
MAXIMUM_DELAY = 10.0
DELAY_MULTIPLIER = 2.0


def _bq_retry(method, *args, **kwargs):
    from google.api_core.exceptions import GoogleAPICallError

    delay = 1.0
    while True:
        try:
            return method(*args, **kwargs)
        except GoogleAPICallError as ex:
            logging.info('run %s with exception %s', method.__name__, ex)
            if not _should_retry(ex):
                raise

        delay = min(delay * 2, 10)
        time.sleep(delay)


def _should_retry(ex):
    if len(ex.errors) == 0:
        return False

    reason = ex.errors[0]['reason']
    return reason in ['backendError', 'rateLimitExceeded', 'internalError']



# -*- coding: utf8 -*-
import os
import click
import requests
import sys

from mali_commands.legit.data_sync import DataSync, InvalidJsonFile
from mali_commands.legit.json_utils import multi_line_json_from_data, newline_json_file_from_json_file, \
    json_data_from_files, normalize_item
from mali_commands.legit import MetadataOperationError
from tqdm import tqdm
from mali_commands.commons import add_to_data_if_not_none, output_result
from mali_commands.legit.data_volume import create_data_volume, with_repo, default_data_volume_path, with_repo_dynamic, \
    map_volume
from mali_commands.options import data_volume_id_argument, no_progressbar_option, processes_option, validate_json
from mali_commands.legit.path_utils import expend_and_validate_path, safe_make_dirs, safe_rm_tree, \
    DestPathEnum, has_moniker, bucket_print_name, enumerate_paths_with_info, AccessDenied
import json


@click.group('data')
def data_commands():
    pass


def __expend_and_validate_path(path, expand_vars=True, validate_path=True, abs_path=True):
    try:
        return expend_and_validate_path(path, expand_vars, validate_path, abs_path)
    except IOError:
        click.echo('Folder not found %s' % path, err=True)
        sys.exit(1)


@data_commands.command('map')
@data_volume_id_argument()
@click.option('--dataPath', required=True)
@click.pass_context
def _cmd_add_data_path(ctx, volumeid, datapath):
    config = map_volume(ctx, volumeid, datapath)

    display_name = config.general_config.get('display_name', 'No display name provided')
    click.echo('Initialized data volume %s (%s)' % (config.volume_id, display_name))


@data_commands.command('create')
@click.option('--displayName', required=True)
@click.option('--description', required=False)
@click.option('--org', required=True)
@click.option('--dataPath', required=True)
@click.option('--bucket')
@click.option('--linked/--embedded', is_flag=True, default=False)
@click.pass_context
def _cmd_create_data_volume(ctx, displayname, description, org, datapath, bucket, linked):
    data = {}

    if org == 'me':
        org = None

    add_to_data_if_not_none(data, displayname, "display_name")
    add_to_data_if_not_none(data, org, "org")
    add_to_data_if_not_none(data, description, "description")
    add_to_data_if_not_none(data, not linked, "embedded")

    expiration = ctx.obj.config.readonly_items('data_volumes').get('expiration')
    if expiration:
        data['expiration'] = expiration

    result = ctx.obj.handle_api(ctx.obj, requests.post, "data_volumes", data)

    data_volume_id = result['id']

    datapath = __expend_and_validate_path(datapath)

    params = {}
    if bucket is not None:
        params['object_store'] = {'bucket_name': bucket}

    create_data_volume(data_volume_id, datapath, linked, displayname, description, **params)

    output_result(ctx, result)


@data_commands.command('config')
@data_volume_id_argument()
@click.option('--edit', is_flag=True)
def edit_config_file(volumeid, edit):
    import subprocess

    path = os.path.join(default_data_volume_path(volumeid), 'config')

    if edit:
        subprocess.call(['edit', path])
        return

    with open(path) as f:
        click.echo(f.read())


@data_commands.command('commit')
@data_volume_id_argument()
@click.option('--message', '-m', required=False)
@click.pass_context
def commit_data_volume(ctx, volumeid, message):
    with with_repo(ctx.obj.config, volumeid) as repo:
        result = repo.commit(message)

        if 'commit_id' not in result:
            click.echo('no changeset detected', err=True)
            sys.exit(1)

        output_result(ctx, result)


def process_moniker_data_path(data_path):
    from six.moves.urllib.parse import urlparse, urlunparse

    if not has_moniker(data_path):
        return data_path

    parts = urlparse(data_path)

    return urlunparse((parts.scheme, parts.netloc, '', '', '', ''))


def __print_transfer_info(repo):
    embedded = repo.data_volume_config.object_store_config.get('embedded')

    if embedded:
        bucket_name = repo.data_volume_config.object_store_config.get('bucket_name')

        if bucket_name:
            click.echo('Transfer files from %s to %s' % (bucket_print_name(repo.data_path), bucket_print_name(bucket_name)))
        else:
            click.echo('Transfer files from %s to MissingLink secure bucket' % (bucket_print_name(repo.data_path),))
    else:
        click.echo('Indexing files from %s' % (bucket_print_name(repo.data_path)))


@data_commands.command('setMetadata')
@click.option('--dataPath', required=True, help='Path to the data')
@click.option('--append/--replace', default=False, help='In case metadata data with the same key already exists, `--append` will not replace it, and `--replace` will. Defaults to `--replace`')
@click.option('--metadata-string', '-ms', multiple=True, type=(str, str), help='string metadata(s) to update. you can provide multiple values in the key value format.')
@click.option('--metadata-num', '-mm', multiple=True, type=(str, int), help='integer metadata(s) to update. you can provide multiple values in key value format.')
@click.option('--metadata-float', '-mf', multiple=True, type=(str, float), help='float metadata(s) to update. you can provide multiple values in key value format.')
@click.option('--metadata-boolean', '-mb', multiple=True, type=(str, bool), help='boolean metadata(s) to update. you can provide multiple values in key value format.')
@click.pass_context
def set_metadata(ctx, datapath, append, metadata_num, metadata_float, metadata_boolean, metadata_string):
    new_values_dict = {}
    for key, value in metadata_num:
        new_values_dict[key] = value
    for key, value in metadata_float:
        new_values_dict[key] = value
    for key, value in metadata_boolean:
        new_values_dict[key] = value
    for key, value in metadata_string:
        new_values_dict[key] = value

    def get_current_metadata(file_path):
        if os.path.isfile(file_path + '.metadata.json'):
            try:
                with open(file_path + '.metadata.json') as f:
                    return json.load(f)
            except Exception:
                return {}
        return {}

    def save_meta(file_path, metadata):
        with open(file_path + '.metadata.json', 'w') as f:
            return json.dump(metadata, f)

    for root, subdirs, files in os.walk(datapath):
        for filename in files:
            if filename.endswith('.metadata.json'):
                continue
            file_path = os.path.join(root, filename)
            cur_meta = get_current_metadata(file_path)
            new_meta = {}
            if append:
                new_meta.update(new_values_dict)
                new_meta.update(cur_meta)
            else:
                new_meta.update(cur_meta)
                new_meta.update(new_values_dict)
            save_meta(file_path, new_meta)
            print('%s meta saved' % file_path)


@data_commands.command('sync')
@data_volume_id_argument()
@click.option('--dataPath', required=True)
@click.option('--commit', required=False)
@processes_option()
@no_progressbar_option()
@click.option('--resume', required=False)
@click.pass_context
def sync_to_data_volume(ctx, volumeid, datapath, commit, processes, no_progressbar, resume):
    data_path = __expend_and_validate_path(datapath, expand_vars=False, validate_path=True)

    repo_data_path = process_moniker_data_path(data_path)

    with with_repo_dynamic(ctx, volumeid, repo_data_path) as repo:
        __repo_validate_data_path(repo, volumeid)

        data_sync = DataSync(ctx, repo, no_progressbar, resume)

        try:
            files_to_upload = data_sync.upload_index_and_metadata(data_path)
        except InvalidJsonFile as ex:
            click.echo('Invalid json file %s' % ex.filename, err=True)
            sys.exit(1)
        except AccessDenied as ex:
            click.echo(str(ex))
            sys.exit(1)

        if files_to_upload is not None:
            total_files_to_upload = len(files_to_upload)

            if total_files_to_upload > 0:
                __print_transfer_info(repo)

                with tqdm(total=total_files_to_upload, desc='Syncing files', unit=' files', ncols=80, disable=no_progressbar) as bar:
                    if processes != -1:
                        repo.data_volume_config.object_store_config['processes'] = processes

                    data_sync.upload_in_batches(files_to_upload, callback=lambda _: bar.update())
            else:
                click.echo('No change detected, nothing to upload (metadata only change).')

        if commit is not None:
            repo.commit(commit)


@data_commands.command('add')
@data_volume_id_argument()
@click.option('--files', '-f', multiple=True)
@click.option('--commit', is_flag=True, required=False)
@processes_option()
@no_progressbar_option()
@click.pass_context
def add_to_data_volume(ctx, volumeid, files, commit, processes, no_progressbar):
    all_files = list(enumerate_paths_with_info(files))
    total_files = len(all_files)

    with tqdm(total=total_files, desc="Adding files", unit=' files', ncols=80, disable=no_progressbar) as bar:
        with with_repo(ctx.obj.config, volumeid) as repo:
            data_sync = DataSync(ctx, repo, no_progressbar)
            if processes != -1:
                repo.data_volume_config.object_store_config['processes'] = processes

            data_sync.upload_in_batches(all_files, total_files, callback=lambda x: bar.update())

            if commit:
                repo.commit(commit)


@data_commands.command('clone')
@data_volume_id_argument()
@click.option('--destFolder', '-d', required=True)
@click.option('--destFile', '-df')
@click.option('--query', '-q', required=False)
@click.option('--delete', is_flag=True, required=False)
@click.option('--batchSize', required=False, default=500)
@processes_option()
@no_progressbar_option()
@click.pass_context
def clone_data(ctx, volumeid, destfolder, destfile, query, delete, batchsize, processes, no_progressbar):
    if delete and (destfolder in ('.', './', '/', os.path.expanduser('~'), '~', '~/')):
        raise click.BadParameter("for protection --dest can't point into current directory while using delete")

    dest_folder = __expend_and_validate_path(destfolder, expand_vars=False, validate_path=False)

    root_dest = DestPathEnum.find_root(destfolder)
    dest_pattern = DestPathEnum.get_dest_path(dest_folder, destfile)

    if delete:
        safe_rm_tree(root_dest)

    safe_make_dirs(root_dest)

    with with_repo_dynamic(ctx, volumeid) as repo:
        data_sync = DataSync(ctx, repo, no_progressbar)
        phase_meta = data_sync.download_all(query, root_dest, dest_pattern, batchsize, processes)

        data_sync.save_metadata(root_dest, phase_meta)


@data_commands.group('metadata')
def metadata_commands():
    pass


def stats_from_json(now, json_data):
    return os.stat_result((
        0,  # mode
        0,  # inode
        0,  # device
        0,  # hard links
        0,  # owner uid
        0,  # gid
        len(json_data),  # size
        0,  # atime
        now,
        now,
    ))


@data_commands.command('query')
@data_volume_id_argument()
@click.option('--query', '-q')
@click.option('--batchSize', required=False, default=500)
@click.pass_context
def query_metadata(ctx, volumeid, query, batchsize):
    def get_all_results():
        def get_next_results():
            current_results, current_total_data_points = repo.metadata.query(
                query, max_results=batchsize, start_index=start_index)

            return list(current_results), current_total_data_points

        total_results = []

        start_index = 0

        results, total_data_points = get_next_results()
        while True:
            total_results.extend([normalize_item(item) for item in results])

            if len(results) != batchsize:
                break

            start_index += len(results)

            results, total_data_points = get_next_results()

        return total_results

    try:
        with with_repo_dynamic(ctx, volumeid) as repo:
            all_results = get_all_results()
    except MetadataOperationError as ex:
        click.echo(str(ex), err=True)
        sys.exit(1)

    output_result(ctx, all_results)


def chunks(l, n):
    result = []
    for item in l:
        result.append(item)

        if len(result) == n:
            yield result
            result = []

    if result:
        yield result


class File2(click.File):
    def convert(self, value, param, ctx):
        value = os.path.expanduser(value)

        return super(File2, self).convert(value, param, ctx)


def __repo_validate_data_path(repo, volume_id):
    if repo.data_path:
        return

    msg = 'Data volume {0} was not mapped on this machine, ' \
          'you should call "mali data map {0} --dataPath [root path of data]" ' \
          'in order to work with the volume locally'.format(volume_id)
    click.echo(msg, err=True)
    sys.exit(1)


@metadata_commands.command('add')
@data_volume_id_argument()
@click.option('--files', '-f', multiple=True)
@click.option('--data', '-d', required=False, callback=validate_json)
@click.option('--dataPoint', '-dp', multiple=True)
@click.option('--dataFile', '-df', required=False, type=File2())
@click.option('--property', '-p', required=False, type=(str, str), multiple=True)
@click.option('--propertyInt', '-pi', required=False, type=(str, int), multiple=True)
@click.option('--propertyFloat', '-pf', required=False, type=(str, float), multiple=True)
@click.option('--update/--replace', is_flag=True, default=True, required=False)
@no_progressbar_option()
@click.pass_context
def add_to_metadata(
        ctx, volumeid, files, data, datapoint, datafile, property, propertyint, propertyfloat, update, no_progressbar):
    def get_per_data_entry():
        data_per_entry = data or {}

        for props in (property, propertyint, propertyfloat):
            if not props:
                continue

            for prop_name, prop_val in props:
                data_per_entry[prop_name] = prop_val

        return data_per_entry

    def get_current_data_file():
        if datafile:
            return newline_json_file_from_json_file(datafile)

        entries = datapoint or []
        entries.extend(files or [])

        __repo_validate_data_path(repo, volumeid)

        data_list = json_data_from_files(repo.data_path, entries, get_per_data_entry())

        return multi_line_json_from_data(data_list)

    with with_repo_dynamic(ctx, volumeid) as repo:
        file_obj = get_current_data_file()
        data_sync = DataSync(ctx, repo, no_progressbar)
        data_sync.upload_and_update_metadata(file_obj)


@data_commands.command('list')
@click.pass_context
def list_data_volumes(ctx):
    projects = ctx.obj.handle_api(ctx.obj, requests.get, 'data_volumes')

    output_result(ctx, projects.get('volumes', []), ['id', 'display_name', 'description', 'org'])



# -*- coding: utf8 -*-
import json
import logging
import os
import uuid
from traceback import format_exception_only
import six
import click
import requests
import yaml
from click import exceptions

from mali_commands.commons import output_result
from .utilities import source_tracking


@click.group('run', help='runs an experiment on a cluster. By defaults run on a local cluster ')
def run_commands():
    pass


def _load_recipe(r_path):
    if not os.path.isfile(r_path):
        return {}
    else:
        print('loading defaults from recipe: %s' % r_path)
        with open(r_path) as f:
            return yaml.safe_load(f)


DEFAULT_RECIPE_PATH = '.ml_recipe.yml'


def _source_tracking_data(path_='.'):
    repo_obj = None
    src_data = {}
    try:
        repo_obj = source_tracking.get_repo(path_)
        src_data = _export_repo_state(repo_obj)
    # noinspection PyBroadException
    except Exception as ex:
        src_data['error'] = export_exception(ex)
    return src_data, repo_obj


def export_exception(ex):
    return ('\n'.join(format_exception_only(ex.__class__, ex))).strip()


def _export_repo_state(repo_obj):
    src_data = {}
    try:
        src_data['branch'] = repo_obj.branch.name
        src_data['remote'] = repo_obj.remote_url
        src_data['sha_local'] = repo_obj.head_sha
        src_data['sha_local_url'] = repo_obj.head_sha_url
        src_data['is_dirty'] = not repo_obj.is_clean
        if repo_obj.has_head:
            commit_data = repo_obj.export_commit(repo_obj.repo.head.commit)
        if commit_data is not None:
            src_data.update(commit_data)
    # noinspection PyBroadException
    except Exception as ex:
        src_data['error'] = export_exception(ex)

    return src_data


def get_tracking_repo(repo_obj):
    repo_path = os.path.join(repo_obj.repo.working_dir, '.ml_tracking_repo')
    if os.path.isfile(repo_path):
        with open(repo_path) as f:
            return f.read().strip()
    # TODO: ADD QUERY to track server
    # return "git@github.com:missinglinkai/sim-test-remote-run.git"
    return None


def _sync_working_dir_if_needed(repo_obj, invocation_id):
    try:
        tracking_repository_url = get_tracking_repo(repo_obj)
        if tracking_repository_url is None:
            return {'error': 'no tracking repository found.'}
        logging.info('There is repository tracking enabled. Tracking to repository: {}'.format(tracking_repository_url))
        source_tracking_repo = source_tracking.GitRepoSyncer.clone_tracking_repo(tracking_repository_url)
        commit_tag = source_tracking.GitRepoSyncer.merge_src_to_tracking_repository(repo_obj.repo, source_tracking_repo, br_tag=invocation_id)
        shadow_repo_obj = source_tracking.get_repo(repo=source_tracking_repo)
        cur_br = source_tracking_repo.active_branch
        source_tracking_repo.git.checkout(commit_tag)
        shadow_repo_obj.refresh()
        src_data = _export_repo_state(shadow_repo_obj)
        source_tracking_repo.git.checkout(cur_br)
        if 'error' not in src_data:
            logging.info('Tracking repository sync completed. This experiment source code is available here: {}'.format(src_data['sha_local_url']))
        else:
            logging.info('Tracking repository sync Failed. The Error was: {}'.format(src_data['error']))
        return src_data
    except Exception as ex:
        ex_txt = export_exception(ex)
        logging.error("Failed to init repository tracking. This experiment may not be tracked. Got: " + ex_txt)
        return {'error': ex_txt}


def parse_env_array_to_dict(env_array):
    if env_array is None:
        return {}
    res = {}
    for env_tuple in env_array:
        key, value = env_tuple.split('=')
        key = key.strip()
        value = value.strip()

        if (value.startswith("'") and value.endswith("'")) or (value.startswith('"') and value.endswith('"')):
            value = value[1:-1]
        res[key] = value
    return res


def _read_input_data_and_load_recipe(kwargs):
    recipe = kwargs.pop('recipe', None)
    env = kwargs.pop('env', [])
    persistent_path = kwargs.pop('persistent_path', [])

    # try:
    persistent_paths = None
    if len(persistent_path) > 0:
        persistent_paths = []
        for host_path, mount_path in persistent_path:
            persistent_paths.append(dict(host_path=host_path, mount_path=mount_path))

    input_data = {
        'org': kwargs.pop('org', None),
        'project': kwargs.pop('project', None),
        'image': kwargs.pop('image', None),
        'git_repo': kwargs.pop('git_repo', None),
        'git_tag': kwargs.pop('git_tag', None),
        'source_dir': kwargs.pop('source_dir', None),
        'command': kwargs.pop('command', None),
        'data_query': kwargs.pop('data_query', None),
        'data_volume': kwargs.pop('data_volume', None),
        'data_use_iterator': kwargs.pop('data_iterator', False),
        'data_dest_folder': kwargs.pop('data_dest', None),
        'gpu': kwargs.pop('gpu', False),
        'output_paths': kwargs.pop('output_paths', []),
        'persistent_paths': persistent_paths,
        'env': parse_env_array_to_dict(env)
    }
    input_data['env'] = json.dumps(input_data.get('env', {}))

    # Data test must reside inside /data as it will reside inside the data folder, until is configurable
    if input_data.get('data_dest_folder') is not None and not input_data['data_dest_folder'].startswith('/data'):
        raise click.BadOptionUsage('`--data_dest` must begin with /data')

    # Load Recipe
    recipe_data = _load_recipe(recipe or DEFAULT_RECIPE_PATH)
    for k, v in recipe_data.items():
        if k in input_data and input_data[k] is None:
            input_data[k] = v

    # Apply Defaults
    input_data['gpu'] = False if input_data['gpu'] is None else input_data['gpu']
    input_data['data_use_iterator'] = False if input_data['data_use_iterator'] is None else input_data['data_use_iterator']

    default_image = 'gw000/keras:1.2.2-cpu' if not input_data.get('gpu') else 'gw000/keras:1.2.2-gpu'
    input_data['image'] = input_data.get('image') or default_image

    if input_data.get('image') is None and input_data.get('command') is None:
        raise exceptions.BadOptionUsage('No command nor image provided')

    return input_data


def _build_src_pointer(input_data):
    if input_data['git_repo'] is None and input_data['git_tag'] is None:
        exp_id = uuid.uuid4().hex[:5]
        src_repo, x = _source_tracking_data(input_data['source_dir'])
        sync_res = _sync_working_dir_if_needed(x, exp_id)
        if 'error' in sync_res:
            logging.error(sync_res['error'])
        else:
            input_data['git_repo'] = sync_res['remote']
            input_data['git_tag'] = sync_res['branch']
        if input_data['git_repo'] is None and input_data['git_tag'] is None:  # still
            raise exceptions.BadOptionUsage('Failed to obtain git point to use fo the experiment. please provide --git-repo, or --source-dir if you have remote tracking enabled')
            # todo: find the current git path and get tracing path from the server if we don't have git point
            # todo better handling of ignoring


def _save_recipe(input_data, save_recipe):
    with open(save_recipe, 'w') as f:
        save_data = {}
        for k, v in input_data.items():
            if v is not None:
                save_data[k] = v
        yaml.safe_dump(save_data, f, default_flow_style=False)
        print('Configuration saved to ', save_recipe)
    return True


@run_commands.command('xp')
@click.pass_context
@click.option('--org', type=str, help='organization to use')
@click.option('--project', type=int if six.PY3 else long)
@click.option('--image', type=str, required=False, help='Docker image to use, defaults to keras 1.2.2 (gw000)')
@click.option('--git-repo', type=str, required=False, help='Git repository to pull the code from')
@click.option('--git-tag', type=str, required=False, help='Git branch/tag for the git repository. defaults to master. The cloned code will be available under `/code`')
@click.option('--source-dir', type=str, required=False, help='source directory for the experiment.')
@click.option('--command', type=str, required=False, help='command to execute')
@click.option('--gpu/--cpu', required=False, default=None, help='Use GPU for this instance. Your image will need to support this as well. Defaults to --cpu')
@click.option('--data-volume', type=str, required=False, help='data volume to clone data from')
@click.option('--data-query', type=str, required=False, help='query to execute on the data volume')
@click.option('--data-dest', type=str, required=False, help='destination folder and format for cloning data. If provided, must begin with /data')
@click.option('--data-iterator', type=bool, required=False, help='When set to True, data will not be cloned before the experiment and the quarry will be available for the SDK iterator')
@click.option('--recipe', '-r', type=str, required=False, help='recipe file. recipe file is yaml file with the `flag: value` that allows you to specify default values for all params for this function')
@click.option('--save-recipe', type=str, required=False, help='Saves a recipe for this call to the target file and quits. Note the default values are not encoded into the recipe')
@click.option('--env', multiple=True, required=False, help='Environment variables to pass for the invocation in key=value format. You can use this flag multiple times')
@click.option('--output-paths', multiple=True, required=False, help='Paths that will be exported to the Data management at the end of the invocation job. The paths will be available to the the running code under `/path_name` by defaults to `/output`')
@click.option('--persistent-path', multiple=True, type=(str, str), required=False,
              help='` --persistent-path /mnt/data /my_data` \n'
                   'Maps a the `/mnt/data` path (or docker volume name) on the hosting server to the `/my_data` folder of the experiment. \n '
                   'Can be specified more than once.\n'
                   'Persistent mappings are given r/w permission.\n'
                   'USE WITH CAUTION!')
@click.option('--times', type=int, required=False, default=1)
def run_experiment(ctx, **kwargs):
    save_recipe = kwargs.pop('save_recipe', False)
    times = kwargs.pop('times', 1)
    input_data = _read_input_data_and_load_recipe(kwargs)

    _build_src_pointer(input_data)
    org = input_data.pop('org')
    project = input_data.pop('project')

    if save_recipe is not None:
        _save_recipe(input_data, save_recipe)
    print(input_data)
    print(times)
    for i in range(times):
        result = ctx.obj.handle_api(ctx.obj, requests.put, '{}/{}/invoke'.format(org, project), input_data)
        output_result(ctx, result, ['ok', 'invocation'])



# -*- coding: utf8 -*-
import logging
import requests
import six
from requests import HTTPError
from .path_utils import remove_moniker, has_moniker


class GCSServiceOperationError(Exception):
    pass


__gcs_service = None
__gcs_credentials = None


class closing_with_condition(object):
    def __init__(self, thing, should_close):
        self.thing = thing
        self.should_close = should_close

    def __enter__(self):
        return self.thing

    def __exit__(self, *exc_info):
        if self.should_close:
            self.thing.close()


def gcs_credentials():
    import google.auth.exceptions

    global __gcs_credentials

    if __gcs_credentials is None:
        try:
            __gcs_credentials, _ = google.auth.default()
        except google.auth.exceptions.DefaultCredentialsError:
            pass

    return __gcs_credentials


def gcs_service(credentials=None):
    from google.cloud import storage

    global __gcs_service

    if __gcs_service is None:
        __gcs_service = storage.Client(credentials=credentials)

    return __gcs_service


class GCSService(object):
    RETRYABLE_ERRORS = (IOError, )
    DEFAULT_MIMETYPE = 'application/octet-stream'
    NUM_RETRIES = 5

    def __init__(self, signed_url_service):
        self.signed_url_service = signed_url_service


class GCSDownload(GCSService):
    def __init__(self, auth_method, signed_url_service):
        super(GCSDownload, self).__init__(signed_url_service)
        self._auth_method = auth_method

    def download(self, object_name):
        auth_session = get_auth_session(self._auth_method)

        signed_urls = self.signed_url_service.get_signed_urls(['GET'], [object_name])
        url = signed_urls['GET'][0]

        r = auth_session.get(url)
        r.raise_for_status()
        data = r.content

        logging.debug('downloaded  %s(%s)', object_name, len(data))

        return data


class GCSDownloadDirectDownload(GCSService):
    def __init__(self):
        super(GCSDownloadDirectDownload, self).__init__(None)

    @classmethod
    def download(cls, bucket_name, object_name):
        bucket_name = remove_moniker(bucket_name)

        gcs = gcs_service()

        blob = gcs.bucket(bucket_name).blob(object_name)
        data = blob.download_as_string()

        return data


class S3DownloadDirectDownload(GCSService):
    def __init__(self):
        super(S3DownloadDirectDownload, self).__init__(None)

    @classmethod
    def download(cls, bucket_name, object_name):
        import boto3

        bucket_name = remove_moniker(bucket_name)

        s3 = boto3.resource('s3')

        obj = s3.Object(bucket_name, object_name)
        try:
            return obj.get()['Body'].read()
        except Exception as ex:
            return None


class GCSUploadDirect(GCSService):
    def __init__(self):
        super(GCSUploadDirect, self).__init__(None)

    @classmethod
    def upload(cls, bucket_name, object_name, file_obj, headers, credentials=None):
        from google.api_core.exceptions import NotFound, PermissionDenied

        gcs = gcs_service(credentials=credentials)

        blob = gcs.bucket(bucket_name).blob(object_name)

        content_type = headers.get('Content-Type')

        try:
            blob.upload_from_file(file_obj, content_type)
        except NotFound:
            raise NonRetryException('bucket %s not found' % bucket_name)
        except PermissionDenied:
            raise NonRetryException('access denied to bucket %s' % bucket_name)


class S3UploadDirect(GCSService):
    def __init__(self):
        super(S3UploadDirect, self).__init__(None)

    @classmethod
    def upload(cls, bucket_name, object_name, file_obj, headers, credentials=None):
        import boto3

        bucket_name = remove_moniker(bucket_name)

        s3 = boto3.client('s3')
        try:
            s3.upload_fileobj(file_obj, bucket_name, object_name)
        except s3.exceptions.NoSuchBucket:
            raise NonRetryException('No such S3 Bucket %s' % bucket_name)

    @classmethod
    def copy(cls, src, dest):
        import boto3
        from botocore.exceptions import ClientError

        s3 = boto3.client('s3')
        try:
            bucket_name, key = dest.split('/', 1)
            s3.copy_object(Bucket=bucket_name, CopySource=src, Key=key)
        except s3.exceptions.NoSuchBucket:
            raise NonRetryException('No such S3 Bucket %s' % bucket_name)
        except ClientError as ex:
            if ex.response.get('Error', {}).get('Code') == 'AccessDenied':
                raise NonRetryException('Access Denied when copying %s' % src)

            raise


class FileWithCallback(object):
    def __init__(self, file_obj, callback):
        file_obj.seek(0, 2)
        self._total = file_obj.tell()
        file_obj.seek(0)

        self._callback = callback
        self._file_obj = file_obj

    def __len__(self):
        return self._total

    def read(self, size):
        data = self._file_obj.read(size)
        if not six.PY2 and isinstance(data, six.string_types):
            data = data.encode()

        if six.PY3 and isinstance(data, six.string_types):
            data = data.encode()

        if self._callback is not None:
            self._callback(len(data))

        return data


class GCSUpload(GCSService):
    def __init__(self, auth_method, head_url, put_url):
        super(GCSUpload, self).__init__(None)
        self._head_url = head_url
        self._put_url = put_url
        self._auth_method = auth_method

    def upload(self, file_obj, headers, callback=None):
        auth_session = get_auth_session(self._auth_method)

        resp = None

        file_obj_with_callback = FileWithCallback(file_obj, callback)

        if self._head_url:
            resp = auth_session.head(self._head_url)

            if resp.status_code in (204, 404):
                logging.debug('file not found, uploading')
                resp = None

        if resp is None:
            resp = auth_session.put(self._put_url, data=file_obj_with_callback, headers=headers)

        if resp.status_code in (401, 403):
            raise NonRetryException()

        resp.raise_for_status()


class GCSDeleteAll(GCSService):
    @classmethod
    def delete_all(cls, bucket_name, volume_id, max_files=None):
        import google.cloud.exceptions

        logging.info('delete all at %s/%s', bucket_name, volume_id)
        gcs = gcs_service()

        try:
            list_iter = gcs.bucket(bucket_name).list_blobs(prefix=volume_id)
        except google.cloud.exceptions.NotFound:
            logging.warning('bucket %s was not found', bucket_name)
            return

        total_deleted = 0
        for blob in list_iter:
            try:
                gcs.bucket(bucket_name).delete_blob(blob.name)
            except google.cloud.exceptions.NotFound:
                pass

            total_deleted += 1

            if max_files is not None and max_files == total_deleted:
                break

        logging .info('total deleted %s', total_deleted)

        return total_deleted


def get_auth_session(auth_method):
    if auth_method == 'gcloud':
        import google.auth.transport.requests

        credentials = gcs_credentials()

        return google.auth.transport.requests.AuthorizedSession(credentials)

    return requests


s3_moniker = 's3://'


class NonRetryException(Exception):
    pass


def __retry_if_retry_possible_error(exception):
    logging.debug('got retry exception (upload/download) %s', exception)

    return not isinstance(exception, NonRetryException)


def _default_retry():
    from retrying import retry

    def decor(f):
        return retry(retry_on_exception=__retry_if_retry_possible_error, wait_exponential_multiplier=50, wait_exponential_max=5000)(f)

    return decor


def do_upload(auth, bucket_name, object_name, full_path_to_data, headers, head_url, put_url, credentials=None, callback=None):
    global __gcs_credentials

    if credentials is not None:
        __gcs_credentials = credentials

    def handle_file_object():
        if hasattr(full_path_to_data, 'read'):
            full_path_to_data.seek(0)
            return full_path_to_data

        return open(full_path_to_data, 'rb')

    @_default_retry()
    def with_retry_to_s3():
        file_obj = handle_file_object()

        S3UploadDirect().upload(bucket_name, object_name, file_obj, headers)

    @_default_retry()
    def with_retry_transfer_s3():
        full_s3_path = remove_moniker(full_path_to_data)
        object_name_with_bucket = remove_moniker(bucket_name) + '/' + object_name
        S3UploadDirect().copy(full_s3_path, object_name_with_bucket)

    @_default_retry()
    def with_retry_to_gcs():
        file_obj = handle_file_object()

        try:
            if put_url:
                GCSUpload(auth, head_url, put_url).upload(file_obj, headers, callback)
                return

            GCSUploadDirect().upload(bucket_name, object_name, file_obj, headers)
        except HTTPError as e:
            if e.response.status_code != 412:  # precondition
                return

            raise

    upload_method = with_retry_to_gcs

    if bucket_name is not None and bucket_name.startswith(s3_moniker):
        if full_path_to_data.startswith(s3_moniker):
            upload_method = with_retry_transfer_s3
        else:
            upload_method = with_retry_to_s3

    upload_method()


def __handle_s3_download_if_needed(bucket_name, object_name):
    @_default_retry()
    def with_retry_s3():
        return S3DownloadDirectDownload().download(bucket_name, object_name)

    if object_name.startswith(s3_moniker):
        object_name = remove_moniker(object_name)
        bucket_name, object_name = object_name.split('/', 1)
        return with_retry_s3(), True
    elif bucket_name is not None and bucket_name.startswith(s3_moniker):
        return with_retry_s3(), True

    return None, False


def do_download(auth, bucket_name, object_name, signed_url_service=None):
    data, handled = __handle_s3_download_if_needed(bucket_name, object_name)

    if handled:
        return data

    @_default_retry()
    def with_retry():
        if signed_url_service:
            return GCSDownload(auth, signed_url_service).download(object_name)

        return GCSDownloadDirectDownload().download(bucket_name, object_name)

    return with_retry()


def do_delete_all(bucket_name, volume_id, max_files):
    return GCSDeleteAll(signed_url_service=None).delete_all(bucket_name, volume_id, max_files)



# -*- coding: utf8 -*-
import logging
from collections import OrderedDict

import functools

from ...gcs_utils import gcs_credentials, do_upload, do_delete_all, do_download, s3_moniker
from ...connection_mixin import ConnectionMixin
from ...dulwich.object_store import BaseObjectStore
from ...dulwich.objects import hex_to_filename, Blob


def return_request_result(response, content):
    return response, content


def multi_process_block(f):
    @functools.wraps(f)
    def wrap(_self, *args, **kwargs):
        multi_process_control = _self.get_multi_process_control()
        try:
            result = f(_self, *args, **kwargs)

            multi_process_control.join()

            return result
        except Exception:
            multi_process_control.terminate()
            raise

    return wrap


class GCSObjectStore(ConnectionMixin, BaseObjectStore):
    def __init__(self, connection, use_multiprocess=True, processes=-1):
        super(GCSObjectStore, self).__init__(connection)
        self.__upload_pool = None
        self._use_multiprocess = use_multiprocess
        self._multi_process_control = None
        self._processes = processes
        self._object_store_auth = connection.data_volume_config.object_store_config.get('auth')
        self.__bucket_name = connection.data_volume_config.object_store_config.get('bucket_name')
        self.__embedded = connection.data_volume_config.embedded
        self.__volume_id = self._connection.data_volume_config.volume_id
        self._signed_url_service = None

    def delete_all(self, max_files=None):
        return do_delete_all(self.__bucket_name, self.__volume_id, max_files)

    @classmethod
    def get_content_headers(cls, content_type=None):

        headers = OrderedDict()
        if content_type:
            headers['Content-Type'] = content_type

        headers['x-goog-acl'] = 'public-read'
        headers['x-goog-if-generation-match'] = '0'

        return headers

    @property
    def processes(self):
        return self._processes if self.is_multiprocess else 1

    @processes.setter
    def processes(self, value):
        self._processes = value

    @property
    def is_multiprocess(self):
        return self._use_multiprocess and self._processes != 1

    def close(self):
        logging.debug('%s closing', self.__class__)
        if self._multi_process_control is not None:
            self._multi_process_control.close()

        logging.debug('%s closed', self.__class__)

    @classmethod
    def _get_shafile_path(cls, sha):
        # Check from object dir
        return hex_to_filename('objects', sha)

    @classmethod
    def on_upload_error(cls, ex):
        raise ex

    def get_multi_process_control(self):
        self.__init_multi_process_if_needed()
        return self._multi_process_control

    def __init_multi_process_if_needed(self):
        if self._multi_process_control is None:
            from ...multi_process_control import get_multi_process_control

            self._multi_process_control = get_multi_process_control(self.processes)

    def upload(self, obj, head_url=None, put_url=None, callback=None):
        object_name = self._get_shafile_path(obj.sha)

        if self.__bucket_name and not self.__bucket_name.startswith(s3_moniker):
            credentials = gcs_credentials()
        else:
            credentials = None

        content_type = obj.content_type
        headers = GCSObjectStore.get_content_headers(content_type)

        object_name = '%s/%s' % (self.__volume_id, object_name)

        args = (
            self._object_store_auth,
            self.__bucket_name, object_name, obj.full_path, headers, head_url, put_url, credentials)

        self._multi_process_control.execute(do_upload, args=args, callback=callback)

    def add_object(self, obj):

        """Add a single object to this object store.

        :param obj: Object to add
        """
        self.upload(obj)

    def _get_loose_object(self, metadata):
        logging.debug('get object %s', metadata)

        sha = metadata['@id']

        if self.__embedded:
            object_name = '%s/%s' % (self.__volume_id, self._get_shafile_path(sha))
        else:
            object_name = metadata['@url']

        data = do_download(
            self._object_store_auth, self.__bucket_name, object_name, signed_url_service=self._signed_url_service)

        blob = Blob()
        blob.set_raw_chunks([data], sha)
        return blob

    def get_raw(self, metadata):
        """Obtain the raw text for an object.

        :param metadata: metadata for the object.
        :return: tuple with numeric type and object contents.
        """
        ret = self._get_loose_object(metadata)
        if ret is not None:
            return ret.type_num, ret.as_raw_string()

        raise KeyError(metadata)

    @property
    def packs(self):
        raise NotImplementedError(self.packs)

    def __iter__(self):
        raise NotImplementedError(self.__iter__)

    @multi_process_block
    def add_objects(self, objects, callback=None):
        for obj in objects:
            self.upload(obj, callback=callback)

    def contains_packed(self, sha):
        raise NotImplementedError(self.contains_packed)

    def contains_loose(self, sha):
        raise NotImplementedError(self.contains_loose)



# -*- coding: utf8 -*-
import mimetypes
import warnings

mimetypes.init()
mimetypes.add_type(mimetypes.types_map.get('.jpg'), '.jfif')


def get_content_type(body):
    import puremagic

    if not body:
        return None

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        try:
            ext = puremagic.from_string(body)
            return mimetypes.types_map.get(ext)
        except puremagic.PureError:
            return None
